#include <iostream>
#include <cstdarg>
#include <string>
#include <map>

#include "autoptr.hh"
#include "dictionary.hh"

using namespace std;

typedef autoptr<class node> nodeptr;

static map<string, class node *> nodes;

class node : public ptrable
{
public:
    nodeptr l, r; /* autoptr automatically initializes these as NULL */
    
    dictitem bits;

public:
    node() : count(0) { bits.bitcount = 0; }
    node(const string &ss, unsigned c) : s(ss), count(c) { bits.bitcount = 0; }
    ~node() { /* autoptr takes care of itself */ }
    
    void judge()
    {
        nodes.clear();
        dumper(false, bits.bitcount);
    }
    void dump()
    {
        dumper(true, bits.bitcount);
    }
    
    unsigned weight() const;
    inline unsigned length() const
    {
        return bits.bitcount;
    }    
    inline unsigned mass() const
    {
        unsigned ssize = s.size(); if(!ssize)ssize = 1;
        unsigned m = count * ssize * length();
        if(l)m += l->mass();
        if(r)m += r->mass();
        return m;
    }
    
    void add(const nodeptr &p);    
    void balance();
    
    unsigned which(const BitBuf &src, unsigned bitpos, string &target) const;

private:
    void dumper(bool dodump=false, unsigned value=0, unsigned bits=0);
    nodeptr *findlessthan(unsigned desired)
    {
        unsigned lweight = 0;
        if(l) { lweight = l->weight(); if(!lweight)l = NULL; }
        
        unsigned rweight = 0;
        if(r) { rweight = r->weight(); if(!rweight)r = NULL; }
        
        if(lweight > desired || !lweight)
        {
            // Jos L oli kelvoton, tarkistetaan R
            
            if(rweight > desired || !rweight)
                return NULL;
            
            // Jos R oli kelvoton, ei kelpaa. Muuten palautetaan R.
            return &r;
        }
       
        // Jos R on parempi kuin L, palautetaan R.
        if(rweight > lweight && rweight <= desired)
            return &r;
        
        // Muutoin L.        
        return &l;
    }
    
    void rebuildlevel(unsigned parentlevel)
    {
        bits.bitcount = parentlevel+1;
        if(l)l->rebuildlevel(bits.bitcount);
        if(r)r->rebuildlevel(bits.bitcount);
    }
    
    void omi(nodeptr kumpi)
    {
        s = kumpi->s;
        count = kumpi->count;
        l = kumpi->l; if(l)l->rebuildlevel(bits.bitcount);
        r = kumpi->r; if(r)r->rebuildlevel(bits.bitcount);
    }
    
    void adoptoi();
    void tyrkkaa();

    string s;          // The string value
    unsigned count;    // The repetition count of string
};

unsigned node::weight() const
{
    // Do not note the string length here!!
    unsigned w = count;
    if(l)w += l->weight();
    if(r)w += r->weight();
    return w;
}
void node::add(const nodeptr &p)
{
    if(!l && !r && count) /* Tyhji ei tyrkit! */
    {
        tyrkkaa();
    }
    
    if(!l) { (l=p)->rebuildlevel(bits.bitcount); return; }
    if(!r) { (r=p)->rebuildlevel(bits.bitcount); return; }
    
    // Listn sinne miss on enemmn tilaa.
    
    if(r->weight() > l->weight())
        l->add(p);
    else
        r->add(p);
}

// Tyrkk(tm) - luo oksan ja siirt itsens sinne - adoption vastakohta.
void node::tyrkkaa()
{
    l = new node(s, count);
    count = 0;
    s = "";
    l->bits.bitcount = bits.bitcount+1;
}

// Adoptoi(tm) - siirt ominaisuudet oksalta itselleen - tyrkkyksen vastakohta.
void node::adoptoi()
{
    if(r && !l) { omi(r); return; }
    if(l && !r) { omi(l); return; }
}
    
void node::balance()
{
    adoptoi();
    
    if(!l && !r)return;
    
    // Add 0-branches. This must be done in the following case:
    //  ROOT
    //   |-- LEFT
    //   |     |-- 150
    //   |     `-- 150
    //   `-- none
    // When this occurs, we must make a 0 branch, to make it possible
    // to move that one 150 branch to the "none" branch. LEFT will then
    // adopt the 150 to itself, and the result is:
    //  ROOT
    //   |-- 150
    //   `-- 150
    
    if(!l)add(new node("", 0));
    else if(!r)add(new node("", 0));

    for(;;)
    {
        l->balance();
        r->balance();
        
        unsigned lweight = l->weight();
        unsigned rweight = r->weight();
    
        if(lweight == rweight)break;

        unsigned isoarvo;   nodeptr *iso;   // pointtereita autopointtereihin
        unsigned pieniarvo; nodeptr *pieni;
           
        if(lweight > rweight)
        {
            isoarvo = lweight;
            pieniarvo = rweight;
            iso = &l;
            pieni = &r;
        }
        else
        {
            isoarvo = rweight;
            pieniarvo = lweight;
            iso = &r;
            pieni = &l;
        }
           
        unsigned ero = (isoarvo - pieniarvo) / 2;
        
        // Find the biggest branch with 1..ero weight, from the bigger one.
        
        nodeptr *uhri = (*iso)->findlessthan(ero);
          
        if(!uhri)
        {
            // Too bad. We're unoptimal now.
            break;
        }
        
        // Found it. Move it to the smaller one.
        // We move it to the beginning part of the
        // smaller, so it must be balanced again.
        
        (*pieni)->add(*uhri);
        
        // The bigger one loses one random branch,
        // so it too must be balanced again. It may
        // have to do some adoptions too.
        
        *uhri = NULL;
        (*iso)->adoptoi();
    }
    
    // Do this again to lose the 0-branches.
    if(l && !l->weight())l = NULL;
    if(r && !r->weight())r = NULL;
}

void node::dumper(bool dodump=false, unsigned bitcount=0, unsigned value=0)
{
    if(dodump) // && count
    {
        unsigned a;
        for(a=bitcount; a>0; --a)
            cerr << ((value >> (a-1)) & 1);
    }
    else if(count)
    {
        bits.value = value;
        bits.bitcount = bitcount;
        
        nodes[s] = this;
    }
        
    if(dodump)
    {
        cerr << ": mass: " << mass();
        cerr << ", length: " << bits.bitcount;
        if(count)cerr << " - \"" << s << "\" " << count << " times, weight=" << weight();
        cerr << endl;
    }
    
    if(l) { l->dumper(dodump, bitcount+1, (value << 1) | 0); }
    if(r) { r->dumper(dodump, bitcount+1, (value << 1) | 1); }
}

unsigned node::which(const BitBuf &src, unsigned bitpos, string &target) const
{
    if(count) /* Onko sisltnode (bitcount on kyll kaikissa jotain, paitsi juuressa) */
    {
        if(src.Read(bits.bitcount, bitpos) == bits.value)
        {
            target = s;
            return bits.bitcount;
        }
    }
    
    if(l) { unsigned tmp = l->which(src, bitpos, target); if(tmp)return tmp; }
    if(r) { unsigned tmp = r->which(src, bitpos, target); if(tmp)return tmp; }
    
    return 0;
};

static class Dictionary
{
    node Tree;
public:
    Dictionary()
    {
//atom:~$ make tavut
//g++ -O3 -ansi -pipe -Wall -W -fomit-frame-pointer -g   tavut.cc   -o tavut
//atom:~$ cat /tmp/SeenLog/*.html|tavut>dictionary.dat
        Initialize(
#if 1
#include "dictionary.dat"
#else
            1, "T",
            2, "estaa",
            1, "n",
            2, " t",
            1, "oi",
            2, "sta",
            1, "mi",
#endif
            0,NULL);
    }
    inline unsigned which(const BitBuf &src, unsigned bitpos, string &target) const
    {
        return Tree.which(src, bitpos, target);
    }
private:
    void Initialize(int n, const char *s ...);
    void AddNode(const string &s, unsigned count)
    {
        Tree.add(new node(s, count));
    }
} Dictionary;

void Dictionary::Initialize(int n, const char *s ...)
{
    va_list ap;
    va_start(ap, s);
    
    while(n)
    {
        // if(!*s)s = "!";
        AddNode(s, n);
        
        n = va_arg(ap, int);
        s = va_arg(ap, const char *);
    }
    va_end(ap);
    
    // unsigned alkup = Tree.mass();
    // cerr << "Massa ennen tasapainotusta: " << alkup << endl;
    Tree.balance();
    // unsigned loppu = Tree.mass();
    // cerr << "Massa tasapainotuksen jlkeen: " << loppu << endl;
    Tree.judge();
    // cerr << "Kutistui siis " << loppu*100/alkup << "%:iin alkuperisest.\n";
    // cerr << "Massaa kytetn vain tilastoissa, ei ollenkaan lajittelussa.\n";
    // Tree.dump();
}

const dictitem *bitsoftoken(const string &s)
{
    map<string, node *>::const_iterator i;
    
    // cerr << "Finding token '" << s << "' - ";
    
    i = nodes.find(s);
    if(i == nodes.end())
    {
        // cerr << "not found\n";
        return NULL;
    }
    // cerr << "found!\n";    
    return &i->second->bits;
}

unsigned matchbytes(const string &s)
{
    map<string, node *>::const_iterator i;
    
    // cerr << "Etsitn '" << s << "'\n";
    
    i = nodes.lower_bound(s);
    
    if(i == nodes.end())
    {
        // cerr << "Ylltys! Ei lytynyt mitn. Otetaan viimeinen.\n";
        --i;
    }
    
    // const string &n = i->first;
    
    for(;;)
    {
        if(i->first == s.substr(0, i->first.size()))
        {
            // cerr << "Hienoa, '" << i->first << "' kelpaa.\n";
            return i->first.size();
        }
        if(i == nodes.begin())break;
        --i;
    }
    
    // cerr << "Lower bound oli '" << n << "', mutta mitn ei lytynyt\n";
    
    return 0;
}

static const unsigned literalbits = 0;

static void LiteralFlush(bool force, string &literal, BitBuf &dest)
{
    if(!literal.size())return;
    if(literal.size() >= (1 << literalbits) || force)
    {
        const dictitem *item2 = bitsoftoken("");
        dest.Add(item2->value, item2->bitcount);
        
        if(item2->bitcount > 1)
            cerr << "Ouch, literal is " << item2->bitcount << " bits. Renegerate dictionary.dat please.\n";
        
        unsigned b = literal.size();
        if(literalbits)
            dest.Add(b-1, literalbits);
        
        // cerr << "Adding literal '" << literal << "' (" << b << " bytes)\n";
        for(unsigned a=0; a<b; ++a)
            dest.Add(literal[a], 8);
        literal = "";
    }
}

BitBuf DictCompress(const string &s) return result
{
    // for(map<string, node *>::reverse_iterator i = nodes.rbegin(); i != nodes.rend(); ++i)
    //     cerr << "Token: '" << i->first << "'\n";
        
    unsigned a, b=s.size();
    
    string literal;
    
    for(a=0; a<b; )
    {
        unsigned n = matchbytes(s.substr(a));
        
        const dictitem *item = bitsoftoken(s.substr(a, n));
        
        if(!item)
        {
            cerr << "HANDLESS! No method to handle the first " << n << " bytes of '" << s.substr(a) << "'\n";
            throw;
        }
        
        if(!n)
        {
            literal += s[a];
            LiteralFlush(false, literal, result);
            n = 1;
        }
        else
        {
            LiteralFlush(true, literal, result);
            // cerr << "Adding token '" << s.substr(a, n) << "' (" << item->bitcount << " bits)\n";
            result.Add(item->value, item->bitcount);
        }
        a += n;
    }
    LiteralFlush(true, literal, result);
/*
    cerr << "Compressed to " << result.size() << " bits: ";
    for(unsigned a=0; a<result.size(); ++a)cerr << result.Read(1, a);
    cerr << endl; */
}

string DictDecompress(const BitBuf &buf) return result
{
    unsigned a, b=buf.size();
/*    
    cerr << "Decompressing " << buf.size() << " bits: ";
    for(unsigned a=0; a<buf.size(); ++a)cerr << buf.Read(1, a);
    cerr << endl;
*/
    for(a=0; a<b; )
    {
        string s;
        unsigned len = Dictionary.which(buf, a, s);
        a += len;
        if(!s.size())
        {
            unsigned litelen = 1;
            if(literalbits)
            {
                litelen += buf.Read(literalbits, a);
                a += literalbits;
            }
            for(unsigned i=0; i<litelen; ++i)
            {
                char c = buf.Read(8, a); a += 8;
                result += c;
            }
        }
        else
            result += s;
    }
}
